'''
Tortick is a Telegram-like channel API for the Dark Web.

Details: https://pypi.org/project/tortick
Git repo: https://codeberg.org/screwery/tortick
'''

__version__ = '0.10.0'
__repository__ = 'https://codeberg.org/screwery/tortick'
__bugtracker__ = 'https://codeberg.org/screwery/tortick/issues'

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import base64
import contextlib
from datetime import datetime
import getpass
import hashlib
import json
import logging
import os
import sqlite3
import subprocess
import sys
import tempfile
import time
import uuid
import re

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from fastapi.responses import Response, JSONResponse
import magic
import stem
import stem.connection
import stem.socket
import uvicorn

if 'TORTICK_DEBUG' not in os.environ:
    sys.tracebacklimit = 0

BINARY_CHUNKSIZE = 1<<14

ONION_VERSION = b'\x03'

RESERVED_ENTIDS = {
    'pubinfo':     '00000000-0000-0000-0000-000000000000',
    'avatar':      '11111111-1111-1111-1111-111111111111',
    'compromised': 'cccccccc-cccc-cccc-cccc-cccccccccccc',
    'frozen':      'ffffffff-ffff-ffff-ffff-ffffffffffff'
    }

OBFS4_BRIDGE_REGEXP = re.compile(
    r'(obfs4\s+' +
    r'(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}):(?P<port>\d{1,5})\s+' +
    r'[0-9A-F]{40}\s+' +
    r'cert=[A-Za-z0-9\+\/]{70}\s+' +
    r'iat-mode=[012])'
    )

def exc_handle(exc):
    logging.exception(exc)
    return (False, exc)

# -----=====| TIME FUNC |=====-----

def time_ts2iso(timestamp):
    return datetime.utcfromtimestamp(timestamp).isoformat()

def time_tsnow():
    return int(datetime.utcnow().timestamp())

def time_isonow():
    return time_ts2iso(time_tsnow())

# -----=====| BASE64 FUNC |=====-----

def b64_encode(data):
    return base64.b64encode(data).decode('ascii')

def b64_decode(data):
    return base64.b64decode(data.encode('ascii'))

# -----=====| CRYPTO FUNC |=====-----

def crypto_gensalt():
    return os.urandom(16)

def crypto_kdf(salt, first_time):
    if first_time:
        passphrase1 = getpass.getpass('Create a passphrase: ')
        if not passphrase1:
            raise ValueError('Passphrase must not be empty')
        passphrase2 = getpass.getpass('Confirm: ')
        if passphrase1 != passphrase2:
            raise ValueError('Passphrases do not match')
    else:
        passphrase1 = getpass.getpass('Enter the passphrase: ')
        if not passphrase1:
            raise ValueError('Passphrase must not be empty')
    passphrase = passphrase1.encode('utf-8')
    key_prepass = hashlib.sha256(passphrase).digest()
    kdf_engine_key = Scrypt(salt=salt, length=44, n=2**16, r=8, p=1)
    key_derivate = kdf_engine_key.derive(key_prepass)
    return {
        'key': key_derivate[:32],
        'nonce': key_derivate[-12:]
        }

def crypto_genkey(derivate, salt):
    secret = os.urandom(32)
    enc_obj = AESGCM(derivate['key'])
    return enc_obj.encrypt(derivate['nonce'], secret, salt)

def crypto_pubkey(privkey, derivate, salt):
    enc_obj = AESGCM(derivate['key'])
    key = enc_obj.decrypt(derivate['nonce'], privkey, salt)
    key_object = ed25519.Ed25519PrivateKey.from_private_bytes(key)
    return key_object.public_key().public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
        )

def crypto_sign(data, privkey, derivate, salt):
    enc_obj = AESGCM(derivate['key'])
    key = enc_obj.decrypt(derivate['nonce'], privkey, salt)
    key_object = ed25519.Ed25519PrivateKey.from_private_bytes(key)
    return key_object.sign(data)

def crypto_verify(data, sig, pubkey):
    key_object = ed25519.Ed25519PublicKey.from_public_bytes(pubkey)
    key_object.verify(sig, data)

# -----=====| TOR FUNC |=====-----

def tor_load_bridges(bridges_path):
    result = None
    if bridges_path is not None:
        with open(bridges_path, 'rt', encoding='utf-8') as bridge_file:
            data = bridge_file.read()
        result = OBFS4_BRIDGE_REGEXP.findall(data)
        if not result:
            raise RuntimeError('No obfs4 bridges are found in the file')
    return result

def tor_generate_control_key():
    password = base64.b64encode(os.urandom(96)).decode('ascii')
    command = ['/bin/tor', '--hash-password', password]
    with subprocess.Popen(command, stdout=subprocess.PIPE) as process:
        result = {
            'pass': password,
            'hash': process.communicate()[0].decode('ascii')[:-1]
            }
        if process.returncode:
            raise RuntimeError('Tor returned non-zero exit code')
    return result

def tor_run(root_dir, bridges_path):
    obfs4proxy = tor_load_bridges(bridges_path)
    data_dir = os.path.join(root_dir, 'data')
    pid_path = os.path.join(root_dir, 'pid')
    control_path = os.path.join(root_dir, 'control.socket')
    socks_path = os.path.join(root_dir, 'socks5.socket')
    control = tor_generate_control_key()
    torrc = f'''
        RunAsDaemon 1
        AvoidDiskWrites 1
        CacheDirectoryGroupReadable 0
        DisableDebuggerAttachment 1
        SafeLogging 1
        TruncateLogFile 1
        DataDirectory {data_dir}
        DataDirectoryGroupReadable 0
        UnixSocksGroupWritable 0
        PidFile {pid_path}
        SocksPort unix:{socks_path} OnionTrafficOnly
        ControlPort 0
        TransPort 0
        ControlSocket unix:{control_path}
        ControlSocketsGroupWritable 0
        HashedControlPassword {control["hash"]}
        CookieAuthentication 0'''
    if obfs4proxy is not None:
        torrc += '''
        UseBridges 1
        ClientTransportPlugin obfs4 exec /bin/obfs4proxy managed'''
        for bridge in obfs4proxy:
            torrc += f'''
        Bridge {bridge[0]}'''
    torrc += '\n'
    subprocess.check_call(['/bin/chmod', '-R', '6700', root_dir])
    with subprocess.Popen(['/bin/tor', '-f', '-'], stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE) as tor_instance:
        tor_instance.communicate(input=torrc.encode('ascii'))
        if tor_instance.returncode:
            raise RuntimeError(f'Tor returned non-zero exit code: {tor_instance.returncode}')
    control_socket = stem.socket.ControlSocketFile(control_path)
    stem.connection.authenticate(control_socket, control['pass'])
    control_socket.send('TAKEOWNERSHIP')
    response = str(control_socket.recv())
    if response != 'OK':
        raise RuntimeError('Cannot handle the Tor instance')
    return {'socket': control_socket, 'socks5': socks_path, 'pid': pid_path}

def tor_check_bootstrap(control_socket, timeout):
    current_tag = ''
    progress = 0
    while timeout:
        control_socket.send('GETINFO status/bootstrap-phase')
        response = str(control_socket.recv())
        details, status = response.split('\n')
        if status != 'OK':
            raise RuntimeError(f'Tor control socket error: {status}')
        fields = details.split(' ', 4)[-3:]
        progress = int(fields[0].split('=')[-1])
        tag = fields[1].split('=')[-1]
        summary = fields[2].split('=', 1)[-1][1:-1]
        if current_tag != tag:
            current_tag = str(tag)
            logging.info('Tor bootstrap %s%% (%s)', progress, summary)
            if progress == 100:
                break
        time.sleep(1)
        timeout -= 1
    if progress < 100:
        raise RuntimeError('Tor bootstrap timeout')

def tor_expand_privkey(privkey):
    result = bytearray(hashlib.sha512(privkey).digest())
    result[0] &= 248
    result[31] &= 127
    result[31] |= 64
    result = bytes(result)
    return result

def tor_generate_hs_seckey(privkey, derivate, salt):
    key_obj = AESGCM(derivate['key'])
    key = key_obj.decrypt(derivate['nonce'], privkey, salt)
    expanded_key = tor_expand_privkey(key)
    return 'ED25519-V3:' + base64.b64encode(expanded_key).decode('ascii')

def tor_onion_checksum(public_key):
    checksum_data = b'.onion checksum' + public_key + ONION_VERSION
    return hashlib.sha3_256(checksum_data).digest()[:2]

def tor_pubkey2onion(pubkey):
    checksum = tor_onion_checksum(pubkey)
    binary_address = pubkey + checksum + ONION_VERSION
    return base64.b32encode(binary_address).decode('ascii').lower()

def tor_start_onion(root_dir, control_socket, private_key, derivate, salt):
    onion_socket = os.path.join(root_dir, f'hs.{uuid.uuid4()}.socket')
    onion_key = tor_generate_hs_seckey(private_key, derivate, salt)
    control_socket.send(f'ADD_ONION {onion_key} Port=80,unix:{onion_socket}')
    response = str(control_socket.recv()).split('\n')
    if response[-1] != 'OK':
        raise RuntimeError('Cannot start onion service')
    service_id = response[0].split('=', 1)[-1]
    pubkey = crypto_pubkey(private_key, derivate, salt)
    predicted_sid = tor_pubkey2onion(pubkey)
    if predicted_sid != service_id:
        raise RuntimeError('.onion address assertion error. Please report this bug to the tortick maintainers')
    return {'servid': service_id, 'socket': onion_socket}

@contextlib.contextmanager
def tor_context(bridges_path):
    tempdir = tempfile.TemporaryDirectory()
    control = tor_run(tempdir.name, bridges_path)
    control['tempdir'] = tempdir.name
    try:
        yield control
    finally:
        tempdir.cleanup()
        control['socket'].close()

# -----=====| MISC FUNC |=====-----

def misc_generate_entid():
    entid = None
    while True:
        entid = str(uuid.uuid4())
        if len(set(entid)) > 2:
            break
    return entid

# -----=====| SQL FUNC |=====-----

def sql_dict_factory(cursor, row):
    result = {}
    for idx, col in enumerate(cursor.description):
        result[col[0]] = row[idx]
    return result

def sql_query(database, query, data, output):
    database.row_factory = sql_dict_factory
    cur = database.cursor()
    if data is None:
        cur.execute(query)
    else:
        cur.execute(query, data)
    if output:
        result = cur.fetchall()
    else:
        database.commit()
        result = None
    cur.close()
    return result

def sql_set_value(value, default):
    return default if value is None else value

def sql_check_singularity(fetchdata):
    result = None
    if len(fetchdata) == 1:
        result = fetchdata[0]
    if len(fetchdata) > 1:
        raise RuntimeError('Corrupted database: duplicated keys')
    return result

def sql_check_existence_singularity(fetchdata):
    if len(fetchdata) == 0:
        raise RuntimeError('No records found')
    if len(fetchdata) > 1:
        raise RuntimeError('Corrupted database: duplicated keys')
    return fetchdata[0]

# -----=====| SQL SERVER FUNC |=====-----

def sql_server_tables(database):
    settings_table = '''
        CREATE TABLE tort_settings (
            key STRING PRIMARY KEY,
            value STRING NOT NULL
            )'''
    feed_table = '''
        CREATE TABLE tort_feed (
            entid STRING PRIMARY KEY,
            publish_date INTEGER NOT NULL,
            last_updated INTEGER NOT NULL,
            entity STRING NOT NULL,
            signature STRING NOT NULL
            )'''
    bin_table = '''
        CREATE TABLE tort_bin (
            entid STRING NOT NULL,
            fragnum INTEGER NOT NULL,
            binary BLOB NOT NULL
            )'''
    sql_query(database, settings_table, None, False)
    sql_query(database, feed_table, None, False)
    sql_query(database, bin_table, None, False)

def sql_server_settings_set(database, key, value):
    query = '''
        INSERT INTO tort_settings (key, value) VALUES (:key, :value)
        ON CONFLICT (key) DO UPDATE SET value = excluded.value'''
    data = {'key': key, 'value': json.dumps(value, separators=(',', ':')).encode('ascii')}
    sql_query(database, query, data, False)

def sql_server_settings_get(database, key):
    query = 'SELECT value FROM tort_settings WHERE [key] = :key'
    data = {'key': key}
    result = sql_query(database, query, data, True)
    result = sql_check_existence_singularity(result)
    return json.loads(result['value'])

def sql_server_salt_privkey(database):
    query_salt = sql_server_settings_get(database, 'salt')
    salt = b64_decode(query_salt)
    query_privkey = sql_server_settings_get(database, 'privkey')
    privkey = b64_decode(query_privkey)
    return {'salt': salt, 'privkey': privkey}

def sql_server_entity_new(database, derivate, entity, publish_date=None,
                          entid=None):
    data = {
        'entid': sql_set_value(entid, misc_generate_entid()),
        'publish_date': sql_set_value(publish_date, time_tsnow()),
        'last_updated': time_tsnow(),
        'entity': entity
        }
    query_salt_privkey = sql_server_salt_privkey(database)
    salt, privkey = query_salt_privkey['salt'], query_salt_privkey['privkey']
    signable = json.dumps(data, separators=(',', ':')).encode('ascii')
    sig = crypto_sign(signable, privkey, derivate, salt)
    data['signature'] = b64_encode(sig)
    data['entity'] = json.dumps(data['entity'], separators=(',', ':'))
    query = '''
        INSERT INTO tort_feed (entid, publish_date, last_updated, entity, signature)
        VALUES (:entid, :publish_date, :last_updated, :entity, :signature)'''
    sql_query(database, query, data, False)
    return data['entid']

def sql_server_entity_change(database, derivate, entid, entity=None,
                             publish_date=None, deletion=False):
    # get old record
    query = 'SELECT publish_date, entity FROM tort_feed WHERE ([entid] = :entid)'
    data = {'entid': entid}
    result = sql_query(database, query, data, True)
    result = sql_check_existence_singularity(result)
    if 'binary' in json.loads(result['entity']) and not deletion:
        raise RuntimeError('Binary cannot be edited')
    data = {
        'entid': entid,
        'publish_date': sql_set_value(publish_date, result['publish_date']),
        'last_updated': time_tsnow(),
        'entity': sql_set_value(entity, json.loads(result['entity']))
        }
    query_salt_privkey = sql_server_salt_privkey(database)
    salt, privkey = query_salt_privkey['salt'], query_salt_privkey['privkey']
    signable = json.dumps(data, separators=(',', ':')).encode('ascii')
    sig = crypto_sign(signable, privkey, derivate, salt)
    data['signature'] = b64_encode(sig)
    data['entity'] = json.dumps(data['entity'], separators=(',', ':'))
    # sql query
    query = '''
        UPDATE tort_feed SET
            [publish_date] = :publish_date,
            [last_updated] = :last_updated,
            [entity] = :entity,
            [signature] = :signature
        WHERE [entid] = :entid'''
    sql_query(database, query, data, False)

def sql_server_entity_delete(database, derivate, entid):
    query = 'DELETE FROM tort_bin WHERE [entid] = :entid'
    data = {'entid': entid}
    sql_query(database, query, data, False)
    if entid == RESERVED_ENTIDS['pubinfo']:
        entity = {'channel_name': '', 'channel_info': {}}
    else:
        entity = {}
    sql_server_entity_change(database, derivate, entid, entity=entity,
                          publish_date=0, deletion=True)

def sql_server_pubinfo_change(database, derivate, channel_name=None,
                              channel_info=None):
    # get old record
    query = 'SELECT entity FROM tort_feed WHERE ([entid] = :entid)'
    data = {'entid': RESERVED_ENTIDS['pubinfo']}
    result = sql_query(database, query, data, True)
    result = sql_check_existence_singularity(result)
    # get old record
    pubinfo = json.loads(result['entity'])
    pubinfo['channel_name'] = sql_set_value(channel_name,
                                         pubinfo['channel_name'])
    pubinfo['channel_info'] = sql_set_value(channel_info,
                                         pubinfo['channel_info'])
    sql_server_entity_change(database, derivate, RESERVED_ENTIDS['pubinfo'],
                          entity=pubinfo)

def sql_server_binary_add(database, derivate, filename, publish_date=None,
                          chunksize=BINARY_CHUNKSIZE):
    entid = misc_generate_entid()
    struct = {}
    size, fragnum = 0, 0
    struct['mime'] = magic.Magic(mime=True).from_file(filename)
    struct['blocksize'] = {'default': int(chunksize), 'last': None}
    blocks = []
    query = '''
        INSERT INTO tort_bin (entid, fragnum, binary)
        VALUES (:entid, :fragnum, :binary)'''
    with open(filename, 'rb') as stream:
        while chunk := stream.read(chunksize):
            blocks.append(b64_encode(hashlib.sha256(chunk).digest()))
            struct['blocksize']['last'] = int(len(chunk))
            size += len(chunk)
            fragnum += 1
            data = {'entid': entid, 'fragnum': fragnum, 'binary': chunk}
            sql_query(database, query, data, False)
    struct['total'] = int(len(blocks))
    struct['size'] = int(size)
    struct['blocks'] = blocks
    publish_date = sql_set_value(publish_date, time_tsnow())
    return sql_server_entity_new( database, derivate, {'binary': struct},
                              publish_date=publish_date, entid=entid)

def sql_server_post_add(database, derivate, contents, publish_date=None):
    publish_date = sql_set_value(publish_date, time_tsnow())
    return sql_server_entity_new(database, derivate, {'post': contents},
                              publish_date=publish_date)

def sql_server_post_change(database, derivate, entid, contents=None,
                           publish_date=None):
    sql_server_entity_change(database, derivate, entid,
                          entity=None if contents is None else {'post': contents},
                          publish_date=publish_date)

def sql_server_entity_get(database, entid):
    query = '''
        SELECT * from tort_feed WHERE ([entid] = :entid) AND
        ([publish_date] < :now)'''
    data = {'entid': entid, 'now': time_tsnow()}
    result = sql_query(database, query, data, True)
    result = sql_check_singularity(result)
    if result is not None:
        result['publish_date'] = time_ts2iso(result['publish_date'])
        result['last_updated'] = time_ts2iso(result['last_updated'])
        result['entity'] = json.loads(result['entity'])
    return result

def sql_server_binary_get(database, entid, fragnum):
    query = '''
        SELECT binary from tort_bin WHERE ([entid] = :entid) AND
        ([fragnum] = :fragnum)'''
    data = {'entid': entid, 'fragnum': fragnum}
    result = sql_query(database, query, data, True)
    result = sql_check_singularity(result)
    return bytes() if result is None else result['binary']

def sql_server_create_database(filename, channel_name, channel_info):
    if os.path.exists(filename):
        raise RuntimeError(f'Cannot use "{filename}": already exists')
    database = sqlite3.connect(filename)
    logging.info('Connected to database: "%s"', filename)
    sql_server_tables(database)
    logging.info('Tables created')
    salt = crypto_gensalt()
    derivate = crypto_kdf(salt, first_time=True)
    privkey = crypto_genkey(derivate, salt)
    logging.info('Private key generated')
    pubkey_str = tor_pubkey2onion(crypto_pubkey(privkey, derivate, salt))
    logging.info('Tortik address: %s', pubkey_str)
    settings = {
        'db_version': [0, 9, 5],
        'salt': b64_encode(salt),
        'privkey': b64_encode(privkey)
        }
    for key, value in settings.items():
        sql_server_settings_set(database, key, value)
    logging.info('Database set up')
    sql_server_entity_new(database, derivate,
                    {'channel_name': channel_name, 'channel_info': channel_info},
                    publish_date=0, entid=RESERVED_ENTIDS['pubinfo'])
    logging.info('Channel name and info set')
    database.close()
    logging.info('Database successfully created: "%s"', filename)

def sql_updates_get(database, last_updated):
    query = '''
        SELECT entid, last_updated from tort_feed
        WHERE ([last_updated] >= :last_updated) AND ([publish_date] < :now)
        ORDER BY [last_updated] ASC LIMIT 100'''
    data = {'last_updated': last_updated, 'now': time_tsnow()}
    result = sql_query(database, query, data, True)
    return {
        'total': len(result),
        'records': {item['entid']: item['last_updated'] for item in result}
        }

# -----=====| SERVER FUNC |=====-----

def api_server(db_path, socket):

    def not_found(request, exc):
        return JSONResponse(content={'error': 'unknown_command'}, status_code=exc.status_code)

    app = FastAPI(exception_handlers={404: not_found})

    @app.exception_handler(RequestValidationError)
    def validation_exception_handler(request, exc):
        return JSONResponse({'error': 'fastapi_exc', 'desc': str(exc)})

    @app.get("/api/v0/info")
    def api_info():
        try:
            database = sqlite3.connect(db_path)
            result = sql_server_entity_get(database, RESERVED_ENTIDS['pubinfo'])
            database.close()
            del result['entid']
        except Exception:
            logging.exception('Internal error')
            result = {'error': 'internal_error'}
        if result is None:
            result = {'error': 'no_pubinfo'}
        return result

    @app.get("/api/v0/entity/{entid}")
    def api_entity(entid: uuid.UUID):
        try:
            database = sqlite3.connect(db_path)
            result = sql_server_entity_get(database, str(entid))
            database.close()
        except Exception:
            logging.exception('Internal error')
            result = {'error': 'internal_error'}
        if result is None:
            result = {'error': 'no_entity'}
        return result

    @app.get("/api/v0/binary/{entid}/{fragnum}",
             responses={200:{"content":{"application/octet-stream":{}}}},
             response_class=Response)
    def api_binary(entid: uuid.UUID, fragnum: int):
        result = bytes()
        try:
            database = sqlite3.connect(db_path)
            result = sql_server_binary_get(database, str(entid), fragnum)
            database.close()
        except Exception:
            logging.exception('Internal error')
            result = {'error': 'internal_error'}
        return Response(content=result, media_type="application/octet-stream")

    @app.get("/api/v0/updates/{last_updated}")
    def api_updates(last_updated: int):
        try:
            database = sqlite3.connect(db_path)
            result = sql_updates_get(database, last_updated)
            database.close()
        except Exception:
            logging.exception('Internal error')
            result = {'error': 'internal_error'}
        return result

    if socket is not None:
        parameters = {'uds': socket, 'server_header': False, 'date_header': False}
    else:
        parameters = {'port': 8000, 'server_header': False, 'date_header': False}
    uvicorn.run(app, **parameters)

@contextlib.contextmanager
def server_cmd_context(db_path):
    database = sqlite3.connect(db_path)
    logging.info('Connected to database: "%s"', db_path)
    # get privkey & salt
    query_salt_privkey = sql_server_salt_privkey(database)
    salt, privkey = query_salt_privkey['salt'], query_salt_privkey['privkey']
    logging.info('Salt & privkey loaded')
    derivate = crypto_kdf(salt, first_time=False)
    try:
        yield {'db': database, 'derv': derivate, 'privkey': privkey, 'salt': salt}
    finally:
        database.close()

def tort_server_run(db_path, bridges_path, test_run):
    if test_run:
        api_server(db_path, None)
        logging.info('Tortik server stopped')
    else:
        with tor_context(bridges_path) as tor_values:
            logging.info('Tor instance started')
            with server_cmd_context(db_path) as ctx:
                onion_data = tor_start_onion(tor_values['tempdir'],
                                                tor_values['socket'], ctx['privkey'],
                                                ctx['derv'], ctx['salt'])
            logging.info('Service ID: %s', onion_data['servid'])
            tor_check_bootstrap(tor_values['socket'], timeout=180)
            logging.info('Connection established')
            api_server(db_path, onion_data['socket'])
            logging.info('Tortik server stopped')
        logging.info('Tor instance stopped')

def tort_server_addpost(db_path, contents, publish_date):
    with server_cmd_context(db_path) as ctx:
        result = sql_server_post_add(ctx['db'], ctx['derv'], contents, publish_date)
    logging.info('Post successfully added: %s', result)

def tort_server_editpost(db_path, entid, contents, publish_date):
    with server_cmd_context(db_path) as ctx:
        sql_server_post_change(ctx['db'], ctx['derv'], entid, contents, publish_date)
    logging.info('Post successfully edited: %s', entid)

def tort_server_addbinary(db_path, filename, publish_date, chunk_size):
    with server_cmd_context(db_path) as ctx:
        result = sql_server_binary_add(ctx['db'], ctx['derv'], filename,
                                 publish_date, chunk_size)
    logging.info('Binary successfully added: %s', result)

def tort_server_delentity(db_path, entid):
    with server_cmd_context(db_path) as ctx:
        sql_server_entity_delete(ctx['db'], ctx['derv'], entid)
    logging.info('Entity successfully deleted: %s', entid)

def tort_server_editpubinfo(db_path, channel_name, channel_info):
    with server_cmd_context(db_path) as ctx:
        sql_server_pubinfo_change(ctx['db'], ctx['derv'], channel_name, channel_info)
    logging.info('Public info successfully edited')

## ------======| PARSER |======------

def create_parser():
    '''
    Create CLI arguments parser
    '''
    default_parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=f'tortick {__version__}: Telegram-like channel API for the Dark Web',
        epilog=f'Bugtracker: {__bugtracker__}'
        )
    default_parser.add_argument('-v', '--version', action='version', version=__version__)
    subparsers = default_parser.add_subparsers(title='Commands', dest='command')
    # new channel parser
    newserv_p = subparsers.add_parser('channel-new', help='Create new channel database')
    newserv_p.add_argument('-n', '--name', required=True, type=str, dest='channel_name', help='Channel name')
    newserv_p.add_argument('-i', '--info', required=True, type=json.loads, dest='channel_info',  help='Channel info in JSON format')
    newserv_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    # edit pubinfo parser
    editpub_p = subparsers.add_parser('channel-edit', help='Edit channel public info')
    editpub_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    editpub_p.add_argument('-n', '--name', type=str, default=None, dest='channel_name', help='Channel name')
    editpub_p.add_argument('-i', '--info', type=json.loads, default=None, dest='channel_info',  help='Channel info in JSON format')
    # run channel
    runserv_p = subparsers.add_parser('channel-run', help='Run channel server')
    runserv_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    runserv_p.add_argument('-b', '--bridges', type=str, default=None, dest='bridges_file', help='Tor obfs4 bridges file (one per line)')
    runserv_p.add_argument('-t', '--test', action='store_true', dest='test', help='Run test server at localhost:8000')
    # new post parser
    newpost_p = subparsers.add_parser('entity-new-post', help='Create new post in the channel')
    newpost_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    newpost_p.add_argument('-t', '--text', required=True, type=str, dest='post_contents', help='Post contents')
    newpost_p.add_argument('-p', '--postpone', type=int, default=None, dest='pub_time', help='Custom publication time (default: now)')
    # new binary parser
    newbin_p = subparsers.add_parser('entity-new-bin', help='Create new binary in the channel')
    newbin_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    newbin_p.add_argument('-f', '--filepath', required=True, type=str, dest='file_name', help='File to upload')
    newbin_p.add_argument('-p', '--postpone', type=int, default=None, dest='pub_time', help='Custom publication time (default: now)')
    newbin_p.add_argument('-c', '--chunksize', type=int, default=BINARY_CHUNKSIZE, dest='chunk_size', help=f'Chunk size in bytes (default: {BINARY_CHUNKSIZE})')
    # edit post parser
    editpost_p = subparsers.add_parser('entity-edit-post', help='Edit post in the channel')
    editpost_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    editpost_p.add_argument('-e', '--entid', required=True, type=lambda x: str(uuid.UUID(x)), dest='entid', help='EntID to edit')
    editpost_p.add_argument('-t', '--text', type=str, default=None, dest='post_contents', help='Post contents')
    editpost_p.add_argument('-p', '--postpone', type=int, default=None, dest='pub_time', help='Custom publication time')
    # delete entity parser
    delent_p = subparsers.add_parser('entity-delete', help='Delete entity (binary/post) from the channel')
    delent_p.add_argument('-d', '--db', required=True, type=str, dest='database_file', help='Channel database file')
    delent_p.add_argument('-e', '--entid', required=True, type=lambda x: str(uuid.UUID(x)), dest='entid', help='EntID to delete')
    return default_parser

# -----=====| MAIN |=====-----

def main():
    logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
    parser = create_parser()
    ns = parser.parse_args(sys.argv[1:])
    if ns.command == 'channel-new':
        sql_server_create_database(ns.database_file, ns.channel_name, ns.channel_info)
    elif ns.command == 'channel-edit':
        tort_server_editpubinfo(ns.database_file, ns.channel_name, ns.channel_info)
    elif ns.command == 'entity-new-post':
        tort_server_addpost(ns.database_file, ns.post_contents, ns.pub_time)
    elif ns.command == 'entity-edit-post':
        tort_server_editpost(ns.database_file, ns.entid, ns.post_contents, ns.pub_time)
    elif ns.command == 'entity-delete':
        tort_server_delentity(ns.database_file, ns.entid)
    elif ns.command == 'entity-new-bin':
        tort_server_addbinary(ns.database_file, ns.file_name, ns.pub_time, ns.chunk_size)
    elif ns.command == 'channel-run':
        tort_server_run(ns.database_file, bridges_path=ns.bridges_file, test_run=ns.test)
    else:
        parser.print_help()

if __name__ == '__main__':
    main()
